from __future__ import annotations

from typing import TYPE_CHECKING

from hypothesis import assume, given
from hypothesis.strategies import DrawFn, booleans, composite, lists, none, sampled_from
from pytest import raises
from sqlalchemy import URL, Column, Integer, MetaData, Table

from utilities.hypothesis import integers, temp_paths, text_ascii, urls
from utilities.postgres import (
    _build_pg_dump,
    _build_pg_restore_or_psql,
    _path_pg_dump,
    _PGDumpFormat,
    _resolve_data_only_and_clean,
    _ResolveDataOnlyAndCleanError,
    pg_dump,
    restore,
)
from utilities.typing import get_literal_elements

if TYPE_CHECKING:
    from pathlib import Path

    from utilities.types import MaybeListStr


@composite
def tables(draw: DrawFn, /) -> list[Table | str]:
    metadata = MetaData()
    names = draw(lists(text_ascii(min_size=1), max_size=5, unique=True))
    tables = [
        Table(n, metadata, Column("id", Integer, primary_key=True)) for n in names
    ]
    return [draw(sampled_from([n, t])) for n, t in zip(names, tables, strict=True)]


class TestPGDump:
    @given(
        url=urls(all_=True), path=temp_paths(), logger=text_ascii(min_size=1) | none()
    )
    async def test_main(self, *, url: URL, path: Path, logger: str | None) -> None:
        _ = await pg_dump(url, path, dry_run=True, logger=logger)

    @given(
        url=urls(all_=True),
        path=temp_paths(),
        format_=sampled_from(get_literal_elements(_PGDumpFormat)),
        jobs=integers(min_value=0) | none(),
        create=booleans(),
        extension=lists(text_ascii(min_size=1)) | none(),
        extension_exc=lists(text_ascii(min_size=1)) | none(),
        schema=lists(text_ascii(min_size=1)) | none(),
        schema_exc=lists(text_ascii(min_size=1)) | none(),
        table=tables() | none(),
        table_exc=tables() | none(),
        inserts=booleans(),
        on_conflict_do_nothing=booleans(),
        role=text_ascii(min_size=1) | none(),
        docker=text_ascii(min_size=1) | none(),
    )
    def test_build(
        self,
        *,
        url: URL,
        path: Path,
        format_: _PGDumpFormat,
        jobs: int | None,
        create: bool,
        extension: MaybeListStr | None,
        extension_exc: MaybeListStr | None,
        schema: list[str] | None,
        schema_exc: list[str] | None,
        table: list[Table | str] | None,
        table_exc: list[Table | str] | None,
        inserts: bool,
        on_conflict_do_nothing: bool,
        role: str | None,
        docker: str | None,
    ) -> None:
        _ = _build_pg_dump(
            url,
            path,
            format_=format_,
            jobs=jobs,
            create=create,
            extension=extension,
            extension_exc=extension_exc,
            schema=schema,
            schema_exc=schema_exc,
            table=table,
            table_exc=table_exc,
            inserts=inserts,
            on_conflict_do_nothing=on_conflict_do_nothing,
            role=role,
            docker=docker,
        )

    @given(path=temp_paths(), format_=sampled_from(get_literal_elements(_PGDumpFormat)))
    def test_path(self, *, path: Path, format_: _PGDumpFormat) -> None:
        path = _path_pg_dump(path, format_=format_)
        assert path.suffix in [".sql", ".pgdump", "", ".tar"]


class TestResolveDataOnlyAndClean:
    @given(data_only=booleans(), clean=booleans())
    def test_main(self, *, data_only: bool, clean: bool) -> None:
        _ = assume(not (data_only and clean))
        _ = _resolve_data_only_and_clean(data_only=data_only, clean=clean)

    def test_erorr(self) -> None:
        with raises(
            _ResolveDataOnlyAndCleanError,
            match="Cannot use '--data-only' and '--clean' together",
        ):
            _ = _resolve_data_only_and_clean(data_only=True, clean=True)


class TestRestore:
    @given(
        url=urls(all_=True), path=temp_paths(), logger=text_ascii(min_size=1) | none()
    )
    async def test_main(self, *, url: URL, path: Path, logger: str | None) -> None:
        _ = await restore(url, path, dry_run=True, logger=logger)

    @given(
        url=urls(all_=True),
        path=temp_paths(),
        psql=booleans(),
        create=booleans(),
        jobs=integers(min_value=0) | none(),
        schema=lists(text_ascii(min_size=1)) | none(),
        schema_exc=lists(text_ascii(min_size=1)) | none(),
        table=tables() | none(),
        role=text_ascii(min_size=1) | none(),
        docker=text_ascii(min_size=1) | none(),
    )
    def test_build(
        self,
        *,
        url: URL,
        path: Path,
        psql: bool,
        create: bool,
        jobs: int | None,
        schema: list[str] | None,
        schema_exc: list[str] | None,
        table: list[Table | str] | None,
        role: str | None,
        docker: str | None,
    ) -> None:
        _ = _build_pg_restore_or_psql(
            url,
            path,
            psql=psql,
            create=create,
            jobs=jobs,
            schema=schema,
            schema_exc=schema_exc,
            table=table,
            role=role,
            docker=docker,
        )



from __future__ import annotations

from collections.abc import Callable, Mapping, Sequence
from collections.abc import Set as AbstractSet
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any, cast, override

import utilities.math
from utilities.functions import is_dataclass_instance
from utilities.iterables import SortIterableError, sort_iterable
from utilities.reprlib import get_repr

if TYPE_CHECKING:
    from utilities.types import Dataclass, Number


def is_equal[T](
    x: Any,
    y: Any,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
    extra: Mapping[type[T], Callable[[T, T], bool]] | None = None,
) -> bool:
    """Check if two objects are equal."""
    if type(x) is type(y):
        # extra
        if extra is not None:
            try:
                cmp = next(cmp for cls, cmp in extra.items() if isinstance(x, cls))
            except StopIteration:
                pass
            else:
                x = cast("T", x)
                y = cast("T", y)
                return cmp(x, y)

        # singletons
        if isinstance(x, int | float):
            y = cast("Number", y)
            return utilities.math.is_equal(x, y, rel_tol=rel_tol, abs_tol=abs_tol)
        if isinstance(x, str):  # else Sequence
            y = cast("str", y)
            return x == y
        if is_dataclass_instance(x):
            y = cast("Dataclass", y)
            x_values = asdict(x)
            y_values = asdict(y)
            return is_equal(x_values, y_values, rel_tol=rel_tol, abs_tol=abs_tol)
        if isinstance(x, Exception):
            return is_equal(x.args, y.args)

        # collections
        if isinstance(x, AbstractSet):
            y = cast("AbstractSet[Any]", y)
            try:
                x_sorted = sort_iterable(x)
                y_sorted = sort_iterable(y)
            except SortIterableError:
                return _is_in(x, y) and _is_in(y, x)
            return is_equal(x_sorted, y_sorted, rel_tol=rel_tol, abs_tol=abs_tol)
        if isinstance(x, Mapping):
            y = cast("Mapping[Any, Any]", y)
            x_keys = set(x)
            y_keys = set(y)
            if not is_equal(x_keys, y_keys, rel_tol=rel_tol, abs_tol=abs_tol):
                return False
            x_values = [x[i] for i in x]
            y_values = [y[i] for i in x]
            return is_equal(x_values, y_values, rel_tol=rel_tol, abs_tol=abs_tol)
        if isinstance(x, Sequence):
            y = cast("Sequence[Any]", y)
            if len(x) != len(y):
                return False
            return all(
                is_equal(x_i, y_i, rel_tol=rel_tol, abs_tol=abs_tol)
                for x_i, y_i in zip(x, y, strict=True)
            )

    if isinstance(x, int | float) and isinstance(y, int | float):
        return utilities.math.is_equal(x, y, rel_tol=rel_tol, abs_tol=abs_tol)

    return (type(x) is type(y)) and (x == y)


def _is_in[T](
    x: AbstractSet[Any],
    y: AbstractSet[Any],
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
    extra: Mapping[type[T], Callable[[T, T], bool]] | None = None,
) -> bool:
    return all(
        any(
            is_equal(x_i, y_i, rel_tol=rel_tol, abs_tol=abs_tol, extra=extra)
            for y_i in y
        )
        for x_i in x
    )


@dataclass(kw_only=True, slots=True)
class IsEqualError(Exception):
    x: Any
    y: Any

    @override
    def __str__(self) -> str:
        return f"Unable to sort {get_repr(self.x)} and {get_repr(self.y)}"  # pragma: no cover


__all__ = ["IsEqualError", "is_equal"]



from __future__ import annotations

from collections.abc import Mapping
from contextlib import suppress
from dataclasses import MISSING, dataclass, field, fields, replace
from typing import TYPE_CHECKING, Any, Literal, assert_never, overload, override

from utilities.errors import ImpossibleCaseError
from utilities.functions import (
    get_class_name,
    is_dataclass_class,
    is_dataclass_instance,
)
from utilities.iterables import (
    OneStrEmptyError,
    OneStrNonUniqueError,
    cmp_nullable,
    one_str,
)
from utilities.operator import is_equal
from utilities.parse import (
    _ParseObjectExtraNonUniqueError,
    _ParseObjectParseError,
    parse_object,
    serialize_object,
)
from utilities.re import ExtractGroupError, extract_group
from utilities.sentinel import Sentinel, sentinel
from utilities.text import (
    BRACKETS,
    LIST_SEPARATOR,
    PAIR_SEPARATOR,
    _SplitKeyValuePairsDuplicateKeysError,
    _SplitKeyValuePairsSplitError,
    split_key_value_pairs,
)
from utilities.types import SupportsLT
from utilities.typing import get_type_hints

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator
    from collections.abc import Set as AbstractSet

    from utilities.types import (
        Dataclass,
        ParseObjectExtra,
        SerializeObjectExtra,
        StrMapping,
        StrStrMapping,
    )


def dataclass_repr[T](
    obj: Dataclass,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
    include: Iterable[str] | None = None,
    exclude: Iterable[str] | None = None,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
    extra: Mapping[type[T], Callable[[T, T], bool]] | None = None,
    defaults: bool = False,
    recursive: bool = False,
) -> str:
    """Repr a dataclass, without its defaults."""
    out: dict[str, str] = {}
    for fld in yield_fields(
        obj, globalns=globalns, localns=localns, warn_name_errors=warn_name_errors
    ):
        if (
            fld.keep(
                include=include,
                exclude=exclude,
                rel_tol=rel_tol,
                abs_tol=abs_tol,
                extra=extra,
                defaults=defaults,
            )
            and fld.repr
        ):
            if recursive:
                if is_dataclass_instance(fld.value):
                    repr_ = dataclass_repr(
                        fld.value,
                        globalns=globalns,
                        localns=localns,
                        warn_name_errors=warn_name_errors,
                        include=include,
                        exclude=exclude,
                        rel_tol=rel_tol,
                        abs_tol=abs_tol,
                        extra=extra,
                        defaults=defaults,
                        recursive=recursive,
                    )
                elif isinstance(fld.value, list):
                    repr_ = [
                        dataclass_repr(
                            v,
                            globalns=globalns,
                            localns=localns,
                            warn_name_errors=warn_name_errors,
                            include=include,
                            exclude=exclude,
                            rel_tol=rel_tol,
                            abs_tol=abs_tol,
                            extra=extra,
                            defaults=defaults,
                            recursive=recursive,
                        )
                        if is_dataclass_instance(v)
                        else repr(v)
                        for v in fld.value
                    ]
                    repr_ = f"[{', '.join(repr_)}]"
                else:
                    repr_ = repr(fld.value)
            else:
                repr_ = repr(fld.value)
            out[fld.name] = repr_
    cls = get_class_name(obj)
    joined = ", ".join(f"{k}={v}" for k, v in out.items())
    return f"{cls}({joined})"


##


def dataclass_to_dict[T](
    obj: Dataclass,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
    include: Iterable[str] | None = None,
    exclude: Iterable[str] | None = None,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
    extra: Mapping[type[T], Callable[[T, T], bool]] | None = None,
    defaults: bool = False,
    final: Callable[[type[Dataclass], StrMapping], StrMapping] | None = None,
    recursive: bool = False,
) -> StrMapping:
    """Convert a dataclass to a dictionary."""
    out: StrMapping = {}
    for fld in yield_fields(
        obj, globalns=globalns, localns=localns, warn_name_errors=warn_name_errors
    ):
        if fld.keep(
            include=include,
            exclude=exclude,
            rel_tol=rel_tol,
            abs_tol=abs_tol,
            extra=extra,
            defaults=defaults,
        ):
            if recursive:
                if is_dataclass_instance(fld.value):
                    value = dataclass_to_dict(
                        fld.value,
                        globalns=globalns,
                        localns=localns,
                        warn_name_errors=warn_name_errors,
                        include=include,
                        exclude=exclude,
                        rel_tol=rel_tol,
                        abs_tol=abs_tol,
                        extra=extra,
                        defaults=defaults,
                        final=final,
                        recursive=recursive,
                    )
                elif isinstance(fld.value, list):
                    value = [
                        dataclass_to_dict(
                            v,
                            globalns=globalns,
                            localns=localns,
                            warn_name_errors=warn_name_errors,
                            include=include,
                            exclude=exclude,
                            rel_tol=rel_tol,
                            abs_tol=abs_tol,
                            extra=extra,
                            defaults=defaults,
                            final=final,
                            recursive=recursive,
                        )
                        if is_dataclass_instance(v)
                        else v
                        for v in fld.value
                    ]
                else:
                    value = fld.value
            else:
                value = fld.value
            out[fld.name] = value
    return out if final is None else final(type(obj), out)


##


def is_nullable_lt[T: SupportsLT](x: T | None, y: T | None, /) -> bool | None:
    """Compare two nullable fields."""
    match cmp_nullable(x, y):
        case 1:
            return False
        case -1:
            return True
        case 0:
            return None
        case _ as never:
            assert_never(never)


##


def mapping_to_dataclass[T: Dataclass](
    cls: type[T],
    mapping: StrMapping,
    /,
    *,
    fields: Iterable[_YieldFieldsClass[Any]] | None = None,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
    head: bool = False,
    case_sensitive: bool = False,
    allow_extra: bool = False,
) -> T:
    """Construct a dataclass from a mapping."""
    if fields is None:
        fields_use = list(
            yield_fields(
                cls,
                globalns=globalns,
                localns=localns,
                warn_name_errors=warn_name_errors,
            )
        )
    else:
        fields_use = fields
    try:
        fields_to_values = str_mapping_to_field_mapping(
            cls,
            mapping,
            fields=fields_use,
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
            head=head,
            case_sensitive=case_sensitive,
            allow_extra=allow_extra,
        )
    except _StrMappingToFieldMappingEmptyError as error:
        raise _MappingToDataClassEmptyError(
            cls=cls, key=error.key, head=head, case_sensitive=case_sensitive
        ) from None
    except _StrMappingToFieldMappingNonUniqueError as error:
        raise _MappingToDataClassNonUniqueError(
            cls=cls,
            key=error.key,
            head=head,
            case_sensitive=case_sensitive,
            first=error.first,
            second=error.second,
        ) from None
    field_names_to_values = {f.name: v for f, v in fields_to_values.items()}
    default = {
        f.name
        for f in fields_use
        if (not isinstance(f.default, Sentinel))
        or (not isinstance(f.default_factory, Sentinel))
    }
    have = set(field_names_to_values) | default
    missing = {f.name for f in fields_use} - have
    if len(missing) >= 1:
        raise _MappingToDataClassMissingValuesError(cls=cls, fields=missing)
    return cls(**field_names_to_values)


@dataclass(kw_only=True, slots=True)
class MappingToDataclassError[T: Dataclass](Exception):
    cls: type[T]


@dataclass(kw_only=True, slots=True)
class _MappingToDataClassEmptyError(MappingToDataclassError):
    key: str
    head: bool = False
    case_sensitive: bool = False

    @override
    def __str__(self) -> str:
        return _empty_error_str(
            self.cls, self.key, head=self.head, case_sensitive=self.case_sensitive
        )


@dataclass(kw_only=True, slots=True)
class _MappingToDataClassNonUniqueError(MappingToDataclassError):
    key: str
    head: bool = False
    case_sensitive: bool = False
    first: str
    second: str

    @override
    def __str__(self) -> str:
        return _non_unique_error_str(
            self.cls,
            self.key,
            self.first,
            self.second,
            head=self.head,
            case_sensitive=self.case_sensitive,
        )


@dataclass(kw_only=True, slots=True)
class _MappingToDataClassMissingValuesError(MappingToDataclassError):
    fields: AbstractSet[str]

    @override
    def __str__(self) -> str:
        desc = ", ".join(map(repr, sorted(self.fields)))
        return f"Unable to construct {get_class_name(self.cls)!r}; missing values for {desc}"


##


def one_field(
    cls: type[Dataclass],
    key: str,
    /,
    *,
    fields: Iterable[_YieldFieldsClass[Any]] | None = None,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
    head: bool = False,
    case_sensitive: bool = False,
) -> _YieldFieldsClass[Any]:
    """Get the unique field a key matches to."""
    if fields is None:
        fields_use = list(
            yield_fields(
                cls,
                globalns=globalns,
                localns=localns,
                warn_name_errors=warn_name_errors,
            )
        )
    else:
        fields_use = fields
    mapping = {f.name: f for f in fields_use}
    try:
        name = one_str(mapping, key, head=head, case_sensitive=case_sensitive)
    except OneStrEmptyError:
        raise _OneFieldEmptyError(
            cls=cls, key=key, head=head, case_sensitive=case_sensitive
        ) from None
    except OneStrNonUniqueError as error:
        raise _OneFieldNonUniqueError(
            cls=cls,
            key=key,
            head=head,
            case_sensitive=case_sensitive,
            first=error.first,
            second=error.second,
        ) from None
    return mapping[name]


@dataclass(kw_only=True, slots=True)
class OneFieldError[T: Dataclass](Exception):
    cls: type[T]
    key: str
    head: bool = False
    case_sensitive: bool = False


@dataclass(kw_only=True, slots=True)
class _OneFieldEmptyError(OneFieldError):
    @override
    def __str__(self) -> str:
        return _empty_error_str(
            self.cls, self.key, head=self.head, case_sensitive=self.case_sensitive
        )


@dataclass(kw_only=True, slots=True)
class _OneFieldNonUniqueError(OneFieldError):
    first: str
    second: str

    @override
    def __str__(self) -> str:
        return _non_unique_error_str(
            self.cls,
            self.key,
            self.first,
            self.second,
            head=self.head,
            case_sensitive=self.case_sensitive,
        )


##


@overload
def replace_non_sentinel(
    obj: Dataclass, /, *, in_place: Literal[True], **kwargs: Any
) -> None: ...
@overload
def replace_non_sentinel[T: Dataclass](
    obj: T, /, *, in_place: Literal[False] = False, **kwargs: Any
) -> T: ...
@overload
def replace_non_sentinel[T: Dataclass](
    obj: T, /, *, in_place: bool = False, **kwargs: Any
) -> T | None: ...
def replace_non_sentinel[T: Dataclass](
    obj: T, /, *, in_place: bool = False, **kwargs: Any
) -> T | None:
    """Replace attributes on a dataclass, filtering out sentinel values."""
    if in_place:
        for k, v in kwargs.items():
            if not isinstance(v, Sentinel):
                setattr(obj, k, v)
        return None
    return replace(
        obj, **{k: v for k, v in kwargs.items() if not isinstance(v, Sentinel)}
    )


##


def serialize_dataclass[T](
    obj: Dataclass,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
    include: Iterable[str] | None = None,
    exclude: Iterable[str] | None = None,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
    extra_equal: Mapping[type[T], Callable[[T, T], bool]] | None = None,
    defaults: bool = False,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    extra_serializers: SerializeObjectExtra | None = None,
) -> str:
    """Serialize a Dataclass."""
    mapping: StrStrMapping = {}
    fields = list(
        yield_fields(
            obj, globalns=globalns, localns=localns, warn_name_errors=warn_name_errors
        )
    )
    for fld in fields:
        if fld.keep(
            include=include,
            exclude=exclude,
            rel_tol=rel_tol,
            abs_tol=abs_tol,
            extra=extra_equal,
            defaults=defaults,
        ):
            mapping[fld.name] = serialize_object(
                fld.value,
                list_separator=list_separator,
                pair_separator=pair_separator,
                extra=extra_serializers,
            )
    return serialize_object(
        mapping,
        list_separator=list_separator,
        pair_separator=pair_separator,
        extra=extra_serializers,
    )


def parse_dataclass[T: Dataclass](
    text_or_mapping: str | StrStrMapping,
    cls: type[T],
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
    head: bool = False,
    case_sensitive: bool = False,
    allow_extra_keys: bool = False,
    extra_parsers: ParseObjectExtra | None = None,
) -> T:
    """Construct a dataclass from a string or a mapping or strings."""
    match text_or_mapping:
        case str() as text:
            keys_to_serializes = _parse_dataclass_split_key_value_pairs(
                text,
                cls,
                list_separator=list_separator,
                pair_separator=pair_separator,
                brackets=brackets,
            )
        case Mapping() as keys_to_serializes:
            ...
        case _ as never:
            assert_never(never)
    fields = list(
        yield_fields(
            cls, globalns=globalns, localns=localns, warn_name_errors=warn_name_errors
        )
    )
    try:
        fields_to_serializes = str_mapping_to_field_mapping(
            cls,
            keys_to_serializes,
            fields=fields,
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
            head=head,
            case_sensitive=case_sensitive,
            allow_extra=allow_extra_keys,
        )
    except _StrMappingToFieldMappingEmptyError as error:
        raise _ParseDataClassStrMappingToFieldMappingEmptyError(
            cls=cls, key=error.key, head=head, case_sensitive=case_sensitive
        ) from None
    except _StrMappingToFieldMappingNonUniqueError as error:
        raise _ParseDataClassStrMappingToFieldMappingNonUniqueError(
            cls=cls,
            key=error.key,
            head=head,
            case_sensitive=case_sensitive,
            first=error.first,
            second=error.second,
        ) from None
    field_names_to_values = {
        f.name: _parse_dataclass_parse_text(
            f,
            t,
            cls,
            list_separator=list_separator,
            pair_separator=pair_separator,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra_parsers,
        )
        for f, t in fields_to_serializes.items()
    }
    try:
        return mapping_to_dataclass(
            cls,
            field_names_to_values,
            fields=fields,
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
            head=head,
            case_sensitive=case_sensitive,
            allow_extra=allow_extra_keys,
        )
    except _MappingToDataClassMissingValuesError as error:
        raise _ParseDataClassMissingValuesError(cls=cls, fields=error.fields) from None


def _parse_dataclass_split_key_value_pairs[T: Dataclass](
    text: str,
    cls: type[T],
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    brackets: Iterable[tuple[str, str]] | None = BRACKETS,
) -> StrStrMapping:
    with suppress(ExtractGroupError):
        text = extract_group(r"^\{?(.*?)\}?$", text)
    try:
        return split_key_value_pairs(
            text,
            list_separator=list_separator,
            pair_separator=pair_separator,
            brackets=brackets,
            mapping=True,
        )
    except _SplitKeyValuePairsSplitError as error:
        raise _ParseDataClassSplitKeyValuePairsSplitError(
            text=error.inner, cls=cls
        ) from None
    except _SplitKeyValuePairsDuplicateKeysError as error:
        raise _ParseDataClassSplitKeyValuePairsDuplicateKeysError(
            cls=cls, counts=error.counts
        ) from None


def _parse_dataclass_parse_text(
    field: _YieldFieldsClass[Any],
    text: str,
    cls: type[Dataclass],
    /,
    *,
    list_separator: str = LIST_SEPARATOR,
    pair_separator: str = PAIR_SEPARATOR,
    head: bool = False,
    case_sensitive: bool = False,
    extra: ParseObjectExtra | None = None,
) -> Any:
    try:
        return parse_object(
            field.type_,
            text,
            list_separator=list_separator,
            pair_separator=pair_separator,
            head=head,
            case_sensitive=case_sensitive,
            extra=extra,
        )
    except _ParseObjectParseError:
        raise _ParseDataClassTextParseError(cls=cls, field=field, text=text) from None
    except _ParseObjectExtraNonUniqueError as error:
        raise _ParseDataClassTextExtraNonUniqueError(
            cls=cls, field=field, first=error.first, second=error.second
        ) from None


@dataclass(kw_only=True, slots=True)
class ParseDataClassError[T: Dataclass](Exception):
    cls: type[T]


@dataclass(kw_only=True, slots=True)
class _ParseDataClassSplitKeyValuePairsSplitError(ParseDataClassError):
    text: str

    @override
    def __str__(self) -> str:
        return f"Unable to construct {get_class_name(self.cls)!r}; failed to split key-value pair {self.text!r}"


@dataclass(kw_only=True, slots=True)
class _ParseDataClassSplitKeyValuePairsDuplicateKeysError(ParseDataClassError):
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return f"Unable to construct {get_class_name(self.cls)!r} since there are duplicate keys; got {self.counts!r}"


@dataclass(kw_only=True, slots=True)
class _ParseDataClassTextParseError(ParseDataClassError):
    field: _YieldFieldsClass[Any]
    text: str

    @override
    def __str__(self) -> str:
        return f"Unable to construct {get_class_name(self.cls)!r} since the field {self.field.name!r} of type {self.field.type_!r} could not be parsed; got {self.text!r}"


@dataclass(kw_only=True, slots=True)
class _ParseDataClassTextExtraNonUniqueError(ParseDataClassError):
    field: _YieldFieldsClass[Any]
    first: type[Any]
    second: type[Any]

    @override
    def __str__(self) -> str:
        return f"Unable to construct {get_class_name(self.cls)!r} since the field {self.field.name!r} of type {self.field.type_!r} must contain exactly one parent class in `extra`; got {self.first!r}, {self.second!r} and perhaps more"


@dataclass(kw_only=True, slots=True)
class _ParseDataClassStrMappingToFieldMappingEmptyError(ParseDataClassError):
    key: str
    head: bool = False
    case_sensitive: bool = False

    @override
    def __str__(self) -> str:
        head = f"Unable to construct {get_class_name(self.cls)!r} since it does not contain"
        tail = _empty_error_str_core(
            self.key, head=self.head, case_sensitive=self.case_sensitive
        )
        return f"{head} {tail}"


@dataclass(kw_only=True, slots=True)
class _ParseDataClassStrMappingToFieldMappingNonUniqueError(ParseDataClassError):
    key: str
    head: bool = False
    case_sensitive: bool = False
    first: str
    second: str

    @override
    def __str__(self) -> str:
        head = f"Unable to construct {get_class_name(self.cls)!r} since it must contain"
        tail = _non_unique_error_str_core(
            self.key,
            self.first,
            self.second,
            head=self.head,
            case_sensitive=self.case_sensitive,
        )
        return f"{head} {tail}"


@dataclass(kw_only=True, slots=True)
class _ParseDataClassMissingValuesError(ParseDataClassError):
    fields: AbstractSet[str]

    @override
    def __str__(self) -> str:
        desc = ", ".join(map(repr, sorted(self.fields)))
        return f"Unable to construct {get_class_name(self.cls)!r}; missing values for {desc}"


##


def str_mapping_to_field_mapping[T: Dataclass, U](
    cls: type[T],
    mapping: Mapping[str, U],
    /,
    *,
    fields: Iterable[_YieldFieldsClass[Any]] | None = None,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
    head: bool = False,
    case_sensitive: bool = False,
    allow_extra: bool = False,
) -> Mapping[_YieldFieldsClass[Any], U]:
    """Convert a string-mapping into a field-mapping."""
    keys_to_fields: Mapping[str, _YieldFieldsClass[Any]] = {}
    for key in mapping:
        try:
            keys_to_fields[key] = one_field(
                cls,
                key,
                fields=fields,
                globalns=globalns,
                localns=localns,
                warn_name_errors=warn_name_errors,
                head=head,
                case_sensitive=case_sensitive,
            )
        except _OneFieldEmptyError:
            if not allow_extra:
                raise _StrMappingToFieldMappingEmptyError(
                    cls=cls, key=key, head=head, case_sensitive=case_sensitive
                ) from None
        except _OneFieldNonUniqueError as error:
            raise _StrMappingToFieldMappingNonUniqueError(
                cls=cls,
                key=key,
                head=head,
                case_sensitive=case_sensitive,
                first=error.first,
                second=error.second,
            ) from None
    return {field: mapping[key] for key, field in keys_to_fields.items()}


@dataclass(kw_only=True, slots=True)
class StrMappingToFieldMappingError[T: Dataclass](Exception):
    cls: type[T]
    key: str
    head: bool = False
    case_sensitive: bool = False


@dataclass(kw_only=True, slots=True)
class _StrMappingToFieldMappingEmptyError(StrMappingToFieldMappingError):
    @override
    def __str__(self) -> str:
        return _empty_error_str(
            self.cls, self.key, head=self.head, case_sensitive=self.case_sensitive
        )


@dataclass(kw_only=True, slots=True)
class _StrMappingToFieldMappingNonUniqueError(StrMappingToFieldMappingError):
    first: str
    second: str

    @override
    def __str__(self) -> str:
        return _non_unique_error_str(
            self.cls,
            self.key,
            self.first,
            self.second,
            head=self.head,
            case_sensitive=self.case_sensitive,
        )


##


@overload
def yield_fields(
    obj: Dataclass,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> Iterator[_YieldFieldsInstance[Any]]: ...
@overload
def yield_fields(
    obj: type[Dataclass],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> Iterator[_YieldFieldsClass[Any]]: ...
def yield_fields(
    obj: Dataclass | type[Dataclass],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> Iterator[_YieldFieldsInstance[Any]] | Iterator[_YieldFieldsClass[Any]]:
    """Yield the fields of a dataclass."""
    if is_dataclass_instance(obj):
        for field in yield_fields(
            type(obj),
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
        ):
            yield _YieldFieldsInstance(
                name=field.name,
                value=getattr(obj, field.name),
                type_=field.type_,
                default=field.default,
                default_factory=field.default_factory,
                init=field.init,
                repr=field.repr,
                hash_=field.hash_,
                compare=field.compare,
                metadata=field.metadata,
                kw_only=field.kw_only,
            )
    elif is_dataclass_class(obj):
        hints = get_type_hints(
            obj, globalns=globalns, localns=localns, warn_name_errors=warn_name_errors
        )
        for field in fields(obj):
            if isinstance(field.type, type):
                type_ = field.type
            else:
                type_ = hints.get(field.name, field.type)
            yield (
                _YieldFieldsClass(
                    name=field.name,
                    type_=type_,
                    default=sentinel if field.default is MISSING else field.default,
                    default_factory=sentinel
                    if field.default_factory is MISSING
                    else field.default_factory,
                    init=field.init,
                    repr=field.repr,
                    hash_=field.hash,
                    compare=field.compare,
                    metadata=dict(field.metadata),
                    kw_only=sentinel if field.kw_only is MISSING else field.kw_only,
                )
            )
    else:
        raise YieldFieldsError(obj=obj)


@dataclass(order=True, unsafe_hash=True, kw_only=True, slots=True)
class _YieldFieldsInstance[T]:
    name: str
    value: T = field(hash=False)
    type_: Any = field(hash=False)
    default: T | Sentinel = field(default=sentinel, hash=False)
    default_factory: Callable[[], T] | Sentinel = field(default=sentinel, hash=False)
    repr: bool = True
    hash_: bool | None = None
    init: bool = True
    compare: bool = True
    metadata: StrMapping = field(default_factory=dict, hash=False)
    kw_only: bool | Sentinel = sentinel

    def equals_default[U](
        self,
        *,
        rel_tol: float | None = None,
        abs_tol: float | None = None,
        extra: Mapping[type[U], Callable[[U, U], bool]] | None = None,
    ) -> bool:
        """Check if the field value equals its default."""
        if isinstance(self.default, Sentinel) and isinstance(
            self.default_factory, Sentinel
        ):
            return False
        if (not isinstance(self.default, Sentinel)) and isinstance(
            self.default_factory, Sentinel
        ):
            expected = self.default
        elif isinstance(self.default, Sentinel) and (
            not isinstance(self.default_factory, Sentinel)
        ):
            expected = self.default_factory()
        else:  # pragma: no cover
            raise ImpossibleCaseError(
                case=[f"{self.default=}", f"{self.default_factory=}"]
            )
        return is_equal(
            self.value, expected, rel_tol=rel_tol, abs_tol=abs_tol, extra=extra
        )

    def keep[U](
        self,
        *,
        include: Iterable[str] | None = None,
        exclude: Iterable[str] | None = None,
        rel_tol: float | None = None,
        abs_tol: float | None = None,
        extra: Mapping[type[U], Callable[[U, U], bool]] | None = None,
        defaults: bool = False,
    ) -> bool:
        """Whether to include a field."""
        if (include is not None) and (self.name not in include):
            return False
        if (exclude is not None) and (self.name in exclude):
            return False
        equal = self.equals_default(rel_tol=rel_tol, abs_tol=abs_tol, extra=extra)
        return (defaults and equal) or not equal


@dataclass(order=True, unsafe_hash=True, kw_only=True, slots=True)
class _YieldFieldsClass[T]:
    name: str
    type_: Any = field(hash=False)
    default: T | Sentinel = field(default=sentinel, hash=False)
    default_factory: Callable[[], T] | Sentinel = field(default=sentinel, hash=False)
    repr: bool = True
    hash_: bool | None = None
    init: bool = True
    compare: bool = True
    metadata: StrMapping = field(default_factory=dict, hash=False)
    kw_only: bool | Sentinel = sentinel


@dataclass(kw_only=True, slots=True)
class YieldFieldsError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Object must be a dataclass instance or class; got {self.obj}"


##


def _empty_error_str(
    cls: type[Dataclass],
    key: str,
    /,
    *,
    head: bool = False,
    case_sensitive: bool = False,
) -> str:
    head_msg = f"Dataclass {get_class_name(cls)!r} does not contain"
    tail_msg = _empty_error_str_core(key, head=head, case_sensitive=case_sensitive)
    return f"{head_msg} {tail_msg}"


def _empty_error_str_core(
    key: str, /, *, head: bool = False, case_sensitive: bool = False
) -> str:
    match head, case_sensitive:
        case False, True:
            return f"a field {key!r}"
        case False, False:
            return f"a field {key!r} (modulo case)"
        case True, True:
            return f"any field starting with {key!r}"
        case True, False:
            return f"any field starting with {key!r} (modulo case)"
        case _ as never:
            assert_never(never)


def _non_unique_error_str(
    cls: type[Dataclass],
    key: str,
    first: str,
    second: str,
    /,
    *,
    head: bool = False,
    case_sensitive: bool = False,
) -> str:
    head_msg = f"Dataclass {get_class_name(cls)!r} must contain"
    tail_msg = _non_unique_error_str_core(
        key, first, second, head=head, case_sensitive=case_sensitive
    )
    return f"{head_msg} {tail_msg}"


def _non_unique_error_str_core(
    key: str,
    first: str,
    second: str,
    /,
    *,
    head: bool = False,
    case_sensitive: bool = False,
) -> str:
    match head, case_sensitive:
        case False, True:
            raise ImpossibleCaseError(  # pragma: no cover
                case=[f"{head=}", f"{case_sensitive=}"]
            )
        case False, False:
            head_msg = f"field {key!r} exactly once (modulo case)"
        case True, True:
            head_msg = f"exactly one field starting with {key!r}"
        case True, False:
            head_msg = f"exactly one field starting with {key!r} (modulo case)"
        case _ as never:
            assert_never(never)
    return f"{head_msg}; got {first!r}, {second!r} and perhaps more"


__all__ = [
    "MappingToDataclassError",
    "OneFieldError",
    "ParseDataClassError",
    "StrMappingToFieldMappingError",
    "YieldFieldsError",
    "dataclass_repr",
    "dataclass_to_dict",
    "is_nullable_lt",
    "mapping_to_dataclass",
    "one_field",
    "parse_dataclass",
    "replace_non_sentinel",
    "str_mapping_to_field_mapping",
    "yield_fields",
]



from __future__ import annotations

from dataclasses import dataclass
from functools import wraps
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, Any, Self, assert_never, cast, override

from eventkit import (
    Constant,
    Count,
    DropWhile,
    Enumerate,
    Event,
    Filter,
    Fork,
    Iterate,
    Map,
    Pack,
    Partial,
    PartialRight,
    Pluck,
    Skip,
    Star,
    Take,
    TakeUntil,
    TakeWhile,
    Timestamp,
)

from utilities.functions import apply_decorators
from utilities.iterables import always_iterable
from utilities.logging import get_logger

if TYPE_CHECKING:
    from collections.abc import Callable

    from utilities.types import Coro, LoggerOrName, MaybeCoro, MaybeIterable, TypeLike


##


def add_listener[E: Event, F: Callable](
    event: E,
    listener: Callable[..., MaybeCoro[None]],
    /,
    *,
    error: Callable[[Event, BaseException], MaybeCoro[None]] | None = None,
    ignore: TypeLike[BaseException] | None = None,
    logger: LoggerOrName | None = None,
    decorators: MaybeIterable[Callable[[F], F]] | None = None,
    done: Callable[..., MaybeCoro[None]] | None = None,
    keep_ref: bool = False,
) -> E:
    """Connect a listener to an event."""
    lifted = lift_listener(
        listener,
        event,
        error=error,
        ignore=ignore,
        logger=logger,
        decorators=decorators,
    )
    return cast("E", event.connect(lifted, done=done, keep_ref=keep_ref))


##


@dataclass(repr=False, kw_only=True)
class LiftedEvent[F: Callable[..., MaybeCoro[None]]]:
    """A lifted version of `Event`."""

    event: Event

    def name(self) -> str:
        return self.event.name()  # pragma: no cover

    def done(self) -> bool:
        return self.event.done()  # pragma: no cover

    def set_done(self) -> None:
        self.event.set_done()  # pragma: no cover

    def value(self) -> Any:
        return self.event.value()  # pragma: no cover

    def connect[F2: Callable](
        self,
        listener: F,
        /,
        *,
        error: Callable[[Event, BaseException], MaybeCoro[None]] | None = None,
        ignore: TypeLike[BaseException] | None = None,
        logger: LoggerOrName | None = None,
        decorators: MaybeIterable[Callable[[F2], F2]] | None = None,
        done: Callable[..., MaybeCoro[None]] | None = None,
        keep_ref: bool = False,
    ) -> Event:
        return add_listener(
            self.event,
            listener,
            error=error,
            ignore=ignore,
            logger=logger,
            decorators=decorators,
            done=done,
            keep_ref=keep_ref,
        )

    def disconnect(
        self, listener: Any, /, *, error: Any = None, done: Any = None
    ) -> Any:
        return self.event.disconnect(  # pragma: no cover
            listener, error=error, done=done
        )

    def disconnect_obj(self, obj: Any, /) -> None:
        self.event.disconnect_obj(obj)  # pragma: no cover

    def emit(self, *args: Any) -> None:
        self.event.emit(*args)  # pragma: no cover

    def emit_threadsafe(self, *args: Any) -> None:
        self.event.emit_threadsafe(*args)  # pragma: no cover

    def clear(self) -> None:
        self.event.clear()  # pragma: no cover

    def run(self) -> list[Any]:
        return self.event.run()  # pragma: no cover

    def pipe(self, *targets: Event) -> Event:
        return self.event.pipe(*targets)  # pragma: no cover

    def fork(self, *targets: Event) -> Fork:
        return self.event.fork(*targets)  # pragma: no cover

    def set_source(self, source: Any, /) -> None:
        self.event.set_source(source)  # pragma: no cover

    def _onFinalize(self, ref: Any) -> None:  # noqa: N802
        self.event._onFinalize(ref)  # noqa: SLF001 # pragma: no cover

    async def aiter(self, *, skip_to_last: bool = False, tuples: bool = False) -> Any:
        async for i in self.event.aiter(  # pragma: no cover
            skip_to_last=skip_to_last, tuples=tuples
        ):
            yield i

    __iadd__ = connect
    __isub__ = disconnect
    __call__ = emit
    __or__ = pipe

    @override
    def __repr__(self) -> str:
        return self.event.__repr__()  # pragma: no cover

    def __len__(self) -> int:
        return self.event.__len__()  # pragma: no cover

    def __bool__(self) -> bool:
        return self.event.__bool__()  # pragma: no cover

    def __getitem__(self, fork_targets: Any, /) -> Fork:
        return self.event.__getitem__(fork_targets)  # pragma: no cover

    def __await__(self) -> Any:
        return self.event.__await__()  # pragma: no cover

    __aiter__ = aiter

    def __contains__(self, c: Any, /) -> bool:
        return self.event.__contains__(c)  # pragma: no cover

    @override
    def __reduce__(self) -> Any:
        return self.event.__reduce__()  # pragma: no cover

    def filter(self, *, predicate: Any = bool) -> Filter:
        return self.event.filter(predicate=predicate)  # pragma: no cover

    def skip(self, *, count: int = 1) -> Skip:
        return self.event.skip(count=count)  # pragma: no cover

    def take(self, *, count: int = 1) -> Take:
        return self.event.take(count=count)  # pragma: no cover

    def takewhile(self, *, predicate: Any = bool) -> TakeWhile:
        return self.event.takewhile(predicate=predicate)  # pragma: no cover

    def dropwhile(self, *, predicate: Any = lambda x: not x) -> DropWhile:  # pyright: ignore[reportUnknownLambdaType]
        return self.event.dropwhile(predicate=predicate)  # pragma: no cover

    def takeuntil(self, notifier: Event, /) -> TakeUntil:
        return self.event.takeuntil(notifier)  # pragma: no cover

    def constant(self, constant: Any, /) -> Constant:
        return self.event.constant(constant)  # pragma: no cover

    def iterate(self, it: Any, /) -> Iterate:
        return self.event.iterate(it)  # pragma: no cover

    def count(self, *, start: int = 0, step: int = 1) -> Count:
        return self.event.count(start=start, step=step)  # pragma: no cover

    def enumerate(self, *, start: int = 0, step: int = 1) -> Enumerate:
        return self.event.enumerate(start=start, step=step)  # pragma: no cover

    def timestamp(self) -> Timestamp:
        return self.event.timestamp()  # pragma: no cover

    def partial(self, *left_args: Any) -> Partial:
        return self.event.partial(*left_args)  # pragma: no cover

    def partial_right(self, *right_args: Any) -> PartialRight:
        return self.event.partial_right(*right_args)  # pragma: no cover

    def star(self) -> Star:
        return self.event.star()  # pragma: no cover

    def pack(self) -> Pack:
        return self.event.pack()  # pragma: no cover

    def pluck(self, *selections: int | str) -> Pluck:
        return self.event.pluck(*selections)  # pragma: no cover

    def map(
        self,
        func: Any,
        /,
        *,
        timeout: float | None = None,
        ordered: bool = True,
        task_limit: int | None = None,
    ) -> Map:
        return self.event.map(  # pragma: no cover
            func, timeout=timeout, ordered=ordered, task_limit=task_limit
        )


##


class TypedEvent[F: Callable[..., MaybeCoro[None]]](Event):
    """A typed version of `Event`."""

    @override
    def connect[F2: Callable](
        self,
        listener: F,
        error: Callable[[Self, BaseException], MaybeCoro[None]] | None = None,
        done: Callable[[Self], MaybeCoro[None]] | None = None,
        keep_ref: bool = False,
        *,
        ignore: TypeLike[BaseException] | None = None,
        logger: LoggerOrName | None = None,
        decorators: MaybeIterable[Callable[[F2], F2]] | None = None,
    ) -> Self:
        lifted = lift_listener(
            listener,
            self,
            error=cast(
                "Callable[[Event, BaseException], MaybeCoro[None]] | None", error
            ),
            ignore=ignore,
            logger=logger,
            decorators=decorators,
        )
        return cast(
            "Self", super().connect(lifted, error=error, done=done, keep_ref=keep_ref)
        )


##


def lift_listener[F1: Callable[..., MaybeCoro[None]], F2: Callable](
    listener: F1,
    event: Event,
    /,
    *,
    error: Callable[[Event, BaseException], MaybeCoro[None]] | None = None,
    ignore: TypeLike[BaseException] | None = None,
    logger: LoggerOrName | None = None,
    decorators: MaybeIterable[Callable[[F2], F2]] | None = None,
) -> F1:
    match error, bool(iscoroutinefunction(listener)):
        case None, False:
            listener_typed = cast("Callable[..., None]", listener)

            @wraps(listener)
            def listener_no_error_sync(*args: Any, **kwargs: Any) -> None:
                try:
                    listener_typed(*args, **kwargs)
                except Exception as exc:  # noqa: BLE001
                    if (ignore is not None) and isinstance(exc, ignore):
                        return
                    get_logger(logger=logger).exception("")

            lifted = listener_no_error_sync

        case None, True:
            listener_typed = cast("Callable[..., Coro[None]]", listener)

            @wraps(listener)
            async def listener_no_error_async(*args: Any, **kwargs: Any) -> None:
                try:
                    await listener_typed(*args, **kwargs)
                except Exception as exc:  # noqa: BLE001
                    if (ignore is not None) and isinstance(exc, ignore):
                        return
                    get_logger(logger=logger).exception("")

            lifted = listener_no_error_async
        case _, _:
            match bool(iscoroutinefunction(listener)), bool(iscoroutinefunction(error)):
                case False, False:
                    listener_typed = cast("Callable[..., None]", listener)
                    error_typed = cast("Callable[[Event, Exception], None]", error)

                    @wraps(listener)
                    def listener_have_error_sync(*args: Any, **kwargs: Any) -> None:
                        try:
                            listener_typed(*args, **kwargs)
                        except Exception as exc:  # noqa: BLE001
                            if (ignore is not None) and isinstance(exc, ignore):
                                return
                            error_typed(event, exc)

                    lifted = listener_have_error_sync
                case False, True:
                    listener_typed = cast("Callable[..., None]", listener)
                    error_typed = cast(
                        "Callable[[Event, Exception], Coro[None]]", error
                    )
                    raise LiftListenerError(listener=listener_typed, error=error_typed)
                case True, _:
                    listener_typed = cast("Callable[..., Coro[None]]", listener)

                    @wraps(listener)
                    async def listener_have_error_async(
                        *args: Any, **kwargs: Any
                    ) -> None:
                        try:
                            await listener_typed(*args, **kwargs)
                        except Exception as exc:  # noqa: BLE001
                            if (ignore is not None) and isinstance(exc, ignore):
                                return None
                            if iscoroutinefunction(error):
                                error_typed = cast(
                                    "Callable[[Event, Exception], Coro[None]]", error
                                )
                                return await error_typed(event, exc)
                            error_typed = cast(
                                "Callable[[Event, Exception], None]", error
                            )
                            error_typed(event, exc)

                    lifted = listener_have_error_async
                case _ as never:
                    assert_never(never)
        case _ as never:
            assert_never(never)

    if decorators is not None:
        lifted = apply_decorators(lifted, *always_iterable(decorators))
    return cast("F1", lifted)


@dataclass(kw_only=True, slots=True)
class LiftListenerError(Exception):
    listener: Callable[..., None]
    error: Callable[[Event, Exception], Coro[None]]

    @override
    def __str__(self) -> str:
        return f"Synchronous listener {self.listener} cannot be paired with an asynchronous error handler {self.error}"


__all__ = [
    "LiftListenerError",
    "LiftedEvent",
    "TypedEvent",
    "add_listener",
    "lift_listener",
]



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from subprocess import CalledProcessError, check_output
from typing import assert_never, override

from libcst import (
    AsName,
    Attribute,
    BaseExpression,
    FormattedString,
    FormattedStringExpression,
    FormattedStringText,
    Import,
    ImportAlias,
    ImportFrom,
    ImportStar,
    Module,
    Name,
)

from utilities.errors import ImpossibleCaseError


def generate_f_string(var: str, suffix: str, /) -> FormattedString:
    """Generate an f-string."""
    return FormattedString([
        FormattedStringExpression(expression=Name(var)),
        FormattedStringText(suffix),
    ])


def generate_import(module: str, /, *, asname: str | None = None) -> Import:
    """Generate an `Import` object."""
    alias = ImportAlias(
        name=split_dotted_str(module), asname=AsName(Name(asname)) if asname else None
    )
    return Import(names=[alias])


def generate_import_from(
    module: str, name: str, /, *, asname: str | None = None
) -> ImportFrom:
    """Generate an `ImportFrom` object."""
    match name, asname:
        case "*", None:
            names = ImportStar()
        case "*", str():
            raise GenerateImportFromError(module=module, asname=asname)
        case _, None:
            alias = ImportAlias(name=Name(name))
            names = [alias]
        case _, str():
            alias = ImportAlias(name=Name(name), asname=AsName(Name(asname)))
            names = [alias]
        case _ as never:
            assert_never(never)
    return ImportFrom(module=split_dotted_str(module), names=names)


@dataclass(kw_only=True, slots=True)
class GenerateImportFromError(Exception):
    module: str
    asname: str | None = None

    @override
    def __str__(self) -> str:
        return f"Invalid import: 'from {self.module} import * as {self.asname}'"


##


@dataclass(kw_only=True, slots=True)
class _ParseImportOutput:
    module: str
    name: str | None = None


def parse_import(import_: Import | ImportFrom, /) -> Sequence[_ParseImportOutput]:
    """Parse an import."""
    match import_:
        case Import():
            return [_parse_import_one(n) for n in import_.names]
        case ImportFrom():
            if (attr_or_name := import_.module) is None:
                raise _ParseImportEmptyModuleError(import_=import_)
            module = join_dotted_str(attr_or_name)
            match import_.names:
                case Sequence() as names:
                    return [_parse_import_from_one(module, n) for n in names]
                case ImportStar():
                    return [_ParseImportOutput(module=module, name="*")]
                case _ as never:
                    assert_never(never)
        case _ as never:
            assert_never(never)


def _parse_import_one(alias: ImportAlias, /) -> _ParseImportOutput:
    return _ParseImportOutput(module=join_dotted_str(alias.name))


def _parse_import_from_one(module: str, alias: ImportAlias, /) -> _ParseImportOutput:
    match alias.name:
        case Name(name):
            return _ParseImportOutput(module=module, name=name)
        case Attribute() as attr:
            raise _ParseImportAliasError(module=module, attr=attr)
        case _ as never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class ParseImportError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _ParseImportEmptyModuleError(ParseImportError):
    import_: ImportFrom

    @override
    def __str__(self) -> str:
        return f"Module must not be None; got {self.import_}"


@dataclass(kw_only=True, slots=True)
class _ParseImportAliasError(ParseImportError):
    module: str
    attr: Attribute

    @override
    def __str__(self) -> str:
        attr = self.attr
        return f"Invalid alias name; got module {self.module!r} and attribute '{attr.value}.{attr.attr}'"


##


def split_dotted_str(dotted: str, /) -> Name | Attribute:
    """Split a dotted string into a name/attribute."""
    parts = dotted.split(".")
    node = Name(parts[0])
    for part in parts[1:]:
        node = Attribute(value=node, attr=Name(part))
    return node


def join_dotted_str(name_or_attr: Name | Attribute, /) -> str:
    """Join a dotted from from a name/attribute."""
    parts: Sequence[str] = []
    curr: BaseExpression | Name | Attribute = name_or_attr
    while True:
        match curr:
            case Name(value=value):
                parts.append(value)
                break
            case Attribute(value=value, attr=Name(value=attr_value)):
                parts.append(attr_value)
                curr = value
            case BaseExpression():  # pragma: no cover
                raise ImpossibleCaseError(case=[f"{curr=}"])
            case _ as never:
                assert_never(never)
    return ".".join(reversed(parts))


##


def render_module(source: str | Module, /) -> str:
    """Render a module."""
    match source:  # skipif-ci
        case str() as text:
            try:
                return check_output(["ruff", "format", "-"], input=text, text=True)
            except CalledProcessError:  # pragma: no cover
                return text
        case Module() as module:
            return render_module(module.code)
        case _ as never:
            assert_never(never)


##


__all__ = [
    "GenerateImportFromError",
    "ParseImportError",
    "generate_f_string",
    "generate_import",
    "generate_import_from",
    "join_dotted_str",
    "parse_import",
    "render_module",
    "split_dotted_str",
]



from __future__ import annotations

from contextlib import nullcontext, suppress
from dataclasses import dataclass
from functools import wraps
from sys import maxsize
from typing import TYPE_CHECKING, override

from pottery import AIORedlock, ExtendUnlockedLock
from pottery.exceptions import ReleaseUnlockedLock
from redis.asyncio import Redis

from utilities.asyncio import loop_until_succeed, sleep_td, timeout_td
from utilities.contextlib import enhanced_async_context_manager
from utilities.contextvars import yield_set_context
from utilities.iterables import always_iterable
from utilities.logging import get_logger
from utilities.whenever import MILLISECOND, SECOND, to_seconds

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable, Iterable
    from contextvars import ContextVar

    from whenever import Delta

    from utilities.types import Coro, LoggerOrName, MaybeIterable

_NUM: int = 1
_TIMEOUT_TRY_ACQUIRE: Delta = SECOND
_TIMEOUT_RELEASE: Delta = 10 * SECOND
_SLEEP: Delta = MILLISECOND


##


async def extend_lock(
    *, lock: AIORedlock | None = None, raise_on_redis_errors: bool | None = None
) -> None:
    """Extend a lock."""
    if lock is not None:
        await lock.extend(raise_on_redis_errors=raise_on_redis_errors)


##


@enhanced_async_context_manager
async def try_yield_coroutine_looper(
    redis: MaybeIterable[Redis],
    key: str,
    /,
    *,
    num: int = _NUM,
    timeout_release: Delta = _TIMEOUT_RELEASE,
    num_extensions: int | None = None,
    timeout_acquire: Delta = _TIMEOUT_TRY_ACQUIRE,
    sleep_acquire: Delta = _SLEEP,
    throttle: Delta | None = None,
    logger: LoggerOrName | None = None,
    sleep_error: Delta | None = None,
    context: ContextVar[bool] | None = None,
) -> AsyncIterator[CoroutineLooper | None]:
    """Try acquire access to a coroutine looper."""
    try:  # skipif-ci-and-not-linux
        async with yield_access(
            redis,
            key,
            num=num,
            timeout_release=timeout_release,
            num_extensions=num_extensions,
            timeout_acquire=timeout_acquire,
            sleep=sleep_acquire,
            throttle=throttle,
        ) as lock:
            looper = CoroutineLooper(lock=lock, logger=logger, sleep=sleep_error)
            if context is None:
                yield looper
            else:
                with yield_set_context(context):
                    yield looper
    except _YieldAccessUnableToAcquireLockError as error:  # skipif-ci-and-not-linux
        if logger is not None:
            get_logger(logger=logger).info("%s", error)
        async with nullcontext():
            yield


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class CoroutineLooper:
    """Looper, guarded by a lock, to repeatedly call a coroutine until it succeeds."""

    lock: AIORedlock
    logger: LoggerOrName | None = None
    sleep: Delta | None = None

    async def __call__[**P](
        self, func: Callable[P, Coro[None]], *args: P.args, **kwargs: P.kwargs
    ) -> bool:
        return await loop_until_succeed(
            lambda: self._make_coro(func, *args, **kwargs),
            logger=self.logger,
            errors=ExtendUnlockedLock,
            sleep=self.sleep,
        )

    def _make_coro[**P](
        self, func: Callable[P, Coro[None]], /, *args: P.args, **kwargs: P.kwargs
    ) -> Coro[None]:
        @wraps(func)
        async def wrapped() -> None:
            await extend_lock(lock=self.lock)
            return await func(*args, **kwargs)

        return wrapped()


##


@enhanced_async_context_manager
async def yield_access(
    redis: MaybeIterable[Redis],
    key: str,
    /,
    *,
    num: int = _NUM,
    timeout_release: Delta = _TIMEOUT_RELEASE,
    num_extensions: int | None = None,
    timeout_acquire: Delta | None = None,
    sleep: Delta = _SLEEP,
    throttle: Delta | None = None,
) -> AsyncIterator[AIORedlock]:
    """Acquire access to a locked resource."""
    if num <= 0:
        raise _YieldAccessNumLocksError(key=key, num=num)
    masters = (  # skipif-ci-and-not-linux
        {redis} if isinstance(redis, Redis) else set(always_iterable(redis))
    )
    locks = [  # skipif-ci-and-not-linux
        AIORedlock(
            key=f"{key}_{i}_of_{num}",
            masters=masters,
            auto_release_time=to_seconds(timeout_release),
            num_extensions=maxsize if num_extensions is None else num_extensions,
        )
        for i in range(1, num + 1)
    ]
    lock: AIORedlock | None = None  # skipif-ci-and-not-linux
    try:  # skipif-ci-and-not-linux
        lock = await _get_first_available_lock(
            key, locks, num=num, timeout=timeout_acquire, sleep=sleep
        )
        yield lock
    finally:  # skipif-ci-and-not-linux
        await sleep_td(throttle)
        if lock is not None:
            with suppress(ReleaseUnlockedLock):
                await lock.release()


async def _get_first_available_lock(
    key: str,
    locks: Iterable[AIORedlock],
    /,
    *,
    num: int = _NUM,
    timeout: Delta | None = None,
    sleep: Delta | None = _SLEEP,
) -> AIORedlock:
    locks = list(locks)  # skipif-ci-and-not-linux
    error = _YieldAccessUnableToAcquireLockError(  # skipif-ci-and-not-linux
        key=key, num=num, timeout=timeout
    )
    async with timeout_td(timeout, error=error):  # skipif-ci-and-not-linux
        while True:
            if (result := await _get_first_available_lock_if_any(locks)) is not None:
                return result
            await sleep_td(sleep)


async def _get_first_available_lock_if_any(
    locks: Iterable[AIORedlock], /
) -> AIORedlock | None:
    for lock in locks:  # skipif-ci-and-not-linux
        if await lock.acquire(blocking=False):
            return lock
    return None  # skipif-ci-and-not-linux


@dataclass(kw_only=True, slots=True)
class YieldAccessError(Exception):
    key: str


@dataclass(kw_only=True, slots=True)
class _YieldAccessNumLocksError(YieldAccessError):
    num: int

    @override
    def __str__(self) -> str:
        return f"Number of locks for {self.key!r} must be positive; got {self.num}"


@dataclass(kw_only=True, slots=True)
class _YieldAccessUnableToAcquireLockError(YieldAccessError):
    num: int
    timeout: Delta | None

    @override
    def __str__(self) -> str:
        return f"Unable to acquire any 1 of {self.num} locks for {self.key!r} after {self.timeout}"  # skipif-ci-and-not-linux


__all__ = [
    "CoroutineLooper",
    "YieldAccessError",
    "extend_lock",
    "try_yield_coroutine_looper",
    "yield_access",
]



from collections.abc import Callable
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path
from typing import Any, ClassVar, assert_never
from uuid import UUID

import typed_settings
from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    booleans,
    data,
    ip_addresses,
    tuples,
    uuids,
)
from pytest import mark, param, raises
from typed_settings import EnvLoader, FileLoader, TomlFormat
from whenever import (
    Date,
    DateDelta,
    DateTimeDelta,
    MonthDay,
    PlainDateTime,
    Time,
    TimeDelta,
    YearMonth,
    ZonedDateTime,
)

from utilities.hypothesis import (
    date_deltas,
    date_time_deltas,
    dates,
    month_days,
    paths,
    plain_datetimes,
    temp_paths,
    text_ascii,
    time_deltas,
    times,
    year_months,
    zoned_datetimes,
)
from utilities.os import temp_environ
from utilities.text import strip_and_dedent
from utilities.typed_settings import (
    ExtendedTSConverter,
    LoadSettingsError,
    load_settings,
)

app_names = text_ascii(min_size=1).map(str.lower)


@dataclass(kw_only=True, slots=True)
class _Case[T]:
    cls: type[T]
    strategy: SearchStrategy[T]
    serialize: Callable[[T], str]


class TestExtendedTSConverter:
    cases: ClassVar[list[_Case]] = [
        _Case(cls=Date, strategy=dates(), serialize=Date.format_common_iso),
        _Case(
            cls=DateDelta,
            strategy=date_deltas(parsable=True),
            serialize=DateDelta.format_common_iso,
        ),
        _Case(
            cls=DateTimeDelta,
            strategy=date_time_deltas(parsable=True),
            serialize=DateTimeDelta.format_common_iso,
        ),
        _Case(cls=IPv4Address, strategy=ip_addresses(v=4), serialize=str),
        _Case(cls=IPv6Address, strategy=ip_addresses(v=6), serialize=str),
        _Case(
            cls=MonthDay, strategy=month_days(), serialize=MonthDay.format_common_iso
        ),
        _Case(
            cls=PlainDateTime,
            strategy=plain_datetimes(),
            serialize=PlainDateTime.format_common_iso,
        ),
        _Case(cls=Time, strategy=times(), serialize=Time.format_common_iso),
        _Case(
            cls=TimeDelta, strategy=time_deltas(), serialize=TimeDelta.format_common_iso
        ),
        _Case(cls=UUID, strategy=uuids(), serialize=str),
        _Case(
            cls=YearMonth, strategy=year_months(), serialize=YearMonth.format_common_iso
        ),
        _Case(
            cls=ZonedDateTime,
            strategy=zoned_datetimes(),
            serialize=ZonedDateTime.format_common_iso,
        ),
    ]

    @given(data=data())
    @mark.parametrize(("cls", "strategy"), [param(c.cls, c.strategy) for c in cases])
    def test_default(
        self, *, data: DataObject, cls: type[Any], strategy: SearchStrategy[Any]
    ) -> None:
        default = data.draw(strategy)

        @dataclass(frozen=True, kw_only=True, slots=True)
        class Settings:
            value: cls = default  # pyright: ignore[reportInvalidTypeForm]

        loaded = typed_settings.load_settings(
            Settings, loaders=[], converter=ExtendedTSConverter()
        )
        assert loaded.value == default

    @given(data=data(), root=temp_paths(), app_name=app_names)
    @mark.parametrize(
        ("cls", "strategy", "serialize"),
        [param(c.cls, c.strategy, c.serialize) for c in cases],
    )
    def test_loaded(
        self,
        *,
        data: DataObject,
        root: Path,
        app_name: str,
        cls: type[Any],
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
    ) -> None:
        default, value = data.draw(tuples(strategy, strategy))

        @dataclass(frozen=True, kw_only=True, slots=True)
        class Settings:
            value: cls = default  # pyright: ignore[reportInvalidTypeForm]

        file = Path(root, "file.toml")
        _ = file.write_text(
            strip_and_dedent(f"""
                [{app_name}]
                value = '{serialize(value)}'
            """)
        )
        loaded = typed_settings.load_settings(
            Settings,
            loaders=[
                FileLoader(formats={"*.toml": TomlFormat(app_name)}, files=[file])
            ],
            converter=ExtendedTSConverter(),
        )
        assert loaded.value == value

    @given(
        root=temp_paths(),
        app_name=app_names,
        env_name=text_ascii(min_size=1).map(lambda text: f"TEST_{text}".upper()),
        env_value=text_ascii(min_size=1),
    )
    def test_path_env_var(
        self, *, root: str, app_name: str, env_name: str, env_value: str
    ) -> None:
        @dataclass(frozen=True, kw_only=True, slots=True)
        class Settings:
            value: Path

        file = Path(root, "file.toml")
        _ = file.write_text(
            strip_and_dedent(f"""
                [{app_name}]
                value = '${env_name}'
            """)
        )
        with temp_environ({env_name: env_value}):
            settings = typed_settings.load_settings(
                Settings,
                loaders=[
                    FileLoader(formats={"*.toml": TomlFormat(app_name)}, files=[file])
                ],
                converter=ExtendedTSConverter(resolve_paths=False),
            )
        expected = Path(env_value)
        assert settings.value == expected

    @given(root=temp_paths(), app_name=app_names, path=paths(), resolve=booleans())
    def test_path_resolution(
        self, *, root: str, app_name: str, path: Path, resolve: bool
    ) -> None:
        @dataclass(frozen=True, kw_only=True, slots=True)
        class Settings:
            value: Path

        file = Path(root, "file.toml")
        _ = file.write_text(
            strip_and_dedent(f"""
                [{app_name}]
                value = '{path!s}'
            """)
        )
        settings = typed_settings.load_settings(
            Settings,
            loaders=[
                FileLoader(formats={"*.toml": TomlFormat(app_name)}, files=[file])
            ],
            converter=ExtendedTSConverter(resolve_paths=resolve),
        )
        match resolve:
            case True:
                expected = Path.cwd().joinpath(path)
            case False:
                expected = Path(path)
            case _ as never:
                assert_never(never)
        assert settings.value == expected


class TestLoadSettings:
    @given(root=temp_paths(), datetime=zoned_datetimes())
    def test_main(self, *, root: Path, datetime: ZonedDateTime) -> None:
        @dataclass(frozen=True, kw_only=True, slots=True)
        class Settings:
            datetime: ZonedDateTime

        file = Path(root, "file.toml")
        _ = file.write_text("")
        _ = file.write_text(
            strip_and_dedent(f"""
                [app_name]
                datetime = '{datetime.format_common_iso()}'
            """)
        )
        settings = load_settings(
            Settings, "app_name", filenames="file.toml", start_dir=root
        )
        assert settings.datetime == datetime

    @given(
        prefix=app_names.map(lambda text: f"TEST_{text}".upper()),
        datetime=zoned_datetimes(),
    )
    def test_loaders(self, *, prefix: str, datetime: ZonedDateTime) -> None:
        key = f"{prefix}__DATETIME"

        @dataclass(frozen=True, kw_only=True, slots=True)
        class Settings:
            datetime: ZonedDateTime

        with temp_environ({key: datetime.format_common_iso()}):
            settings = load_settings(
                Settings, "app_name", loaders=[EnvLoader(prefix=f"{prefix}__")]
            )
        assert settings.datetime == datetime

    @mark.parametrize("app_name", [param("app_"), param("app1"), param("app__name")])
    def test_error(self, *, app_name: str) -> None:
        @dataclass(frozen=True, kw_only=True, slots=True)
        class Settings: ...

        with raises(LoadSettingsError, match="Invalid app name; got '.+'"):
            _ = load_settings(Settings, app_name)
